@@ -1,25 +1,30 @@
 """
 Implementation of the renderer.
 """
+import html
 import re
 from dataclasses import dataclass, field
 from typing import Any, List, Union
 
 from paxter.core import (
     CharLoc, Command, Fragment, FragmentList, Identifier, Number,
-    Operator, Text, Token, TokenList,
+    Operator, ShortSymbol, Text, Token, TokenList,
 )
 from paxter.core.exceptions import PaxterRenderError
-from paxter.pyauthor.funcs import flatten
+from paxter.pyauthor.funcs.document import Document, Paragraph
+from paxter.pyauthor.funcs.standards import flatten
 from paxter.pyauthor.wrappers import BaseApply, NormalApply
 
 BACKSLASH_NEWLINE_RE = re.compile(r'\\\n')
+PARAGRAPH_SPLIT_RE = re.compile(
+    r'(?:[ \t\r\f\v]+|(?<!\\))\n(?:[ \t\r\f\v]*\n)+[ \t\r\f\v]*',
+)
 
 
 @dataclass
-class RenderContext:
+class BaseRenderContext:
     """
-    A suite of Paxter document tree renderer.
+    Base rendering class for Paxter parsed document tree.
 
     Users of this renderer may embed and run pyauthor code
     directly from within the Paxter document source file.
@@ -33,17 +38,14 @@ class RenderContext:
     #: Parsed document tree
     tree: FragmentList
 
-    #: Whether the list should always be joined into string
-    #: whenever possible
-    is_joined: bool = True
-
     #: Result of the rendering
     rendered: Union[str, list] = field(init=False)
 
     def __post_init__(self):
-        self.rendered = self.transform_fragment_list(self.tree)
-        if self.is_joined:
-            self.rendered = flatten(self.rendered, is_joined=True)
+        self.rendered = self.render()
+
+    def render(self):
+        return self.transform_fragment_list(self.tree)
 
     def transform_token(self, token: Token) -> Any:
         if isinstance(token, Fragment):
@@ -56,18 +58,20 @@ class RenderContext:
             return self.transform_operator(token)
         if isinstance(token, Number):
             return self.transform_number(token)
+        if isinstance(token, FragmentList):
+            return self.transform_fragment_list(token)
         raise PaxterRenderError(
             "unrecognized token at %(pos)s",
             pos=CharLoc(self.input_text, token.start_pos),
         )
 
     def transform_fragment(self, fragment: Fragment) -> Any:
-        if isinstance(fragment, FragmentList):
-            return self.transform_fragment_list(fragment)
         if isinstance(fragment, Text):
             return self.transform_text(fragment)
         if isinstance(fragment, Command):
             return self.transform_command(fragment)
+        if isinstance(fragment, ShortSymbol):
+            return self.transform_symbol_command(fragment)
         raise PaxterRenderError(
             "unrecognized fragment at %(pos)s",
             pos=CharLoc(self.input_text, fragment.start_pos),
@@ -103,8 +107,6 @@ class RenderContext:
             fragment for fragment in transformed_fragments
             if fragment is not None
         ]
-        if self.is_joined:
-            result = ''.join(str(fragment) for fragment in result)
         return result
 
     def transform_text(self, token: Text) -> str:
@@ -114,37 +116,37 @@ class RenderContext:
         return text
 
     def transform_command(self, token: Command):
-        # Try to evaluate the intro section
-        # using the evaluator function from _intro_eval_
+        # Try to evaluate the starter section
+        # using the evaluator function from _starter_eval_
         try:
-            intro_eval = self.env['_intro_eval_']
+            starter_eval = self.env['_starter_eval_']
         except KeyError as exc:
             raise PaxterRenderError(
-                "expected '_intro_eval_' to be defined at %(pos)s",
+                "expected '_starter_eval_' to be defined at %(pos)s",
                 pos=CharLoc(self.input_text, token.start_pos),
             ) from exc
         try:
-            intro_value = intro_eval(token.intro, self.env)
+            starter_value = starter_eval(token.starter, self.env)
         except PaxterRenderError:
             raise
         except Exception as exc:
             raise PaxterRenderError(
-                "paxter command intro evaluation error at %(pos)s: "
-                f"{token.intro!r}",
+                "paxter command starter evaluation error at %(pos)s: "
+                f"{token.starter!r}",
                 pos=CharLoc(self.input_text, token.start_pos),
             ) from exc
 
-        # Bail out if options section and main arg section are empty
-        if token.options is None and token.main_arg is None:
-            return intro_value
+        # Bail out if option section and main arg section are empty
+        if token.option is None and token.main_arg is None:
+            return starter_value
 
         # Wrap the function if not yet wrapped
-        if not isinstance(intro_value, BaseApply):
-            intro_value = NormalApply(intro_value)
+        if not isinstance(starter_value, BaseApply):
+            starter_value = NormalApply(starter_value)
 
         # Make the call to the wrapped function
         try:
-            return intro_value.call(self, token)
+            return starter_value.call(self, token)
         except PaxterRenderError:
             raise
         except Exception as exc:
@@ -152,3 +154,93 @@ class RenderContext:
                 "paxter apply evaluation error at %(pos)s",
                 pos=CharLoc(self.input_text, token.start_pos),
             ) from exc
+
+    def transform_symbol_command(self, token: ShortSymbol):
+        # Lookup _symbols_ for the desired symbol
+        try:
+            symbols = self.env['_symbols_']
+        except KeyError as exc:
+            raise PaxterRenderError(
+                "expected '_symbols_' to be defined at %(pos)s",
+                pos=CharLoc(self.input_text, token.start_pos),
+            ) from exc
+        try:
+            return symbols[token.symbol]
+        except KeyError as exc:
+            raise PaxterRenderError(
+                f"undefined symbol {token.symbol} within '_symbol_' at %(pos)s",
+                pos=CharLoc(self.input_text, token.start_pos),
+            ) from exc
+        except Exception as exc:
+            raise RuntimeError("unexpected error from within library") from exc
+
+
+@dataclass
+class StringRenderContext(BaseRenderContext):
+    """
+    String-based variant of rendering class.
+    Each fragment list will be flattened into a single string
+    each time it is evaluated.
+    """
+
+    def transform_fragment_list(self, seq: FragmentList) -> str:
+        result = super().transform_fragment_list(seq)
+        result = flatten(result, is_joined=True)
+        return result
+
+
+@dataclass
+class DocumentRenderContext(BaseRenderContext):
+    """
+    Document-based variant of rendering class.
+    This is for constructing HTML or LaTeX documents.
+    """
+    rendered: Document = field(init=False)
+
+    def render(self):
+        # Top-level paragraph splitting
+        document = []  # list of paragraphs
+        paragraph = []  # list of pieces
+        for fragment in self.tree.children:
+            if isinstance(fragment, Text):
+                text = fragment.inner
+                pieces = [
+                    html.escape(BACKSLASH_NEWLINE_RE.sub('', piece))
+                    for piece in PARAGRAPH_SPLIT_RE.split(text)
+                ]
+                if pieces[0].strip():
+                    paragraph.append(pieces[0])
+                if len(pieces) >= 2:
+                    document.append(paragraph)
+                    for piece in pieces[1:-1]:
+                        document.append([piece])
+                    paragraph = []
+                    if pieces[-1].strip():
+                        paragraph.append(pieces[-1])
+            else:
+                paragraph.append(fragment)
+        if paragraph:
+            document.append(paragraph)
+
+        # Re-iterate: for each paragraph in the document
+        # if it is not a list of one non-text token,
+        # then wrap it under paragraph data class.
+        rendered_document = []
+        for paragraph in document:
+            rendered_paragraph = [
+                self.transform_fragment(fragment)
+                if isinstance(fragment, Fragment)
+                else fragment
+                for fragment in paragraph
+            ]
+            if len(paragraph) == 1 and isinstance(paragraph[0], Fragment):
+                rendered_document.append(rendered_paragraph[0])
+            else:
+                rendered_document.append(Paragraph(rendered_paragraph))
+
+        return Document(children=rendered_document)
+
+    def transform_text(self, token: Text) -> str:
+        text = super().transform_text(token)
+        text = html.escape(text)
+        return text
